﻿# coding: utf-8

# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ##### END GPL LICENSE BLOCK #####

bl_info = {
    'name': 'Жесты мышью',
    'author': 'chromoly, доработан www.programishka.ru',
    'version': (0, 2, 6),
    'blender': (2, 5, 6),
    'api': 35853,
    'location': 'View3D > Mouse',
    'description': '',
    'warning': '',
    'category': '3D View'}



import math
from types import MethodType

import bpy
from bpy.props import *
from mathutils import Vector
import bgl
import blf


class MGItem:
    def __init__(self, name='',
                 check=lambda self, context, event, action: True,
                 op=lambda self, context: None,
                 **kw):
       
        self.name = name  # type: str
        self.check = check  # type: function
        self.op = op  # type: function
        for k, v in kw.items():
            setattr(self, k, v)

    def _ch_get(self):
        return self._ch
    def _ch_set(self, ch):
        self._ch= MethodType(ch, self)
    check = property(_ch_get, _ch_set)

    def _op_get(self):
        return self._op
    def _op_set(self, op):
        self._op = MethodType(op, self)
    op = property(_op_get, _op_set)


### Customize if you need #####################################################
path_threthold = 5.0  # update interval
path_max_length = 2000

help_str = \
'''
Дейстивя мышью движение '''+chr(8593)+''' - Вверх, '''+chr(8595)+''' - Вниз, '''+chr(8592)+''' - Влево, '''+chr(8594)+''' - Вправо.
Помощь:  '''+chr(8593)+chr(8595)+chr(8592)+chr(8593)+''', '''+chr(8595)+chr(8593)+chr(8594)+chr(8595)+''': - Вывести помощь на экран.

Сменить тип жестов мыши (Абсолютное движение):
  '''+chr(8595)+''' - (Пробел или Tab): Трансформирование
  '''+chr(8593)+''' - (Пробел or Tab): Вид

Трансормирование (Относительное движение):
  '''+chr(8593)+''' - Захватить/Переместить
  '''+chr(8593)+chr(8594)+'''*, '''+chr(8593)+chr(8592)+'''* - Вращать
  '''+chr(8593)+chr(8595)+'''* - Масштабировать

Вид (Абсолютное движение):
  '''+chr(8595)+''' - Вид спереди (Numpad 1)
  '''+chr(8595)+chr(8593)+''' - Виз сзади (Ctrl Numpad 1)
  '''+chr(8594)+''' - Вид справа (Numpad 3)
  '''+chr(8594)+chr(8592)+''' - Вид слева (Ctrl Numpad 3)
  '''+chr(8593)+''' - Вид сверху (Numpad 7)
  '''+chr(8593)+chr(8595)+''' - Вид снизу (Ctrl Numpad 7)
  '''+chr(8592)+''' - Вид из камеры (Numpad 0)
  '''+chr(8592)+chr(8594)+''' - Вид перспектива/отрография  (Numpad 5)
  '''+chr(8592)+chr(8593)+''' - Вид всё (Home)
  '''+chr(8595)+chr(8594)+''' - Вид глобальный/локальный (Numpad /)
  '''+chr(8592)+chr(8595)+''' - Вид выделенное (Numpad .)
  '''+chr(8592)+chr(8595)+chr(8592)+''', '''+chr(8594)+chr(8595)+''' - Вид выделенное (сохранить масштаб)
  '''+chr(8592)+chr(8595)+chr(8594)+'''*, '''+chr(8593)+chr(8592)+chr(8595)+chr(8594)+'''* - Центрировать вид на 3D курсор (Ctrl Numpad .)
'''


def generate_mgdict():

    mgdict = {}

    ### Transform ###
    transitems = []
    # Translate
    def check(self, context, event, action):
        return len(action) <= 1
    def op(self, context):
        return bpy.ops.transform.translate('INVOKE_REGION_WIN')
    transitems.append(MGItem('Захватить/Переместить', check, op))

    # Rotate
    def check(self, context, event, action):
        if action.startswith(('UR', 'UL')):
            return True
        else:
            return False
    def op(self, context):
        return bpy.ops.transform.rotate('INVOKE_REGION_WIN')
    transitems.append(MGItem('Вращение', check, op))

    # Resize
    def check(self, context, event, action):
        if action.startswith('UD'):
            return True
        else:
            return False
    def op(self, context):
        return bpy.ops.transform.resize('INVOKE_REGION_WIN')
    transitems.append(MGItem('Масштаб', check, op))

    mgdict['transform'] = {'items': transitems, 'relative': True}

    ### View ###
    viewitems = []
    # View
    for name, vtype, ac in (('Вид спереди', 'FRONT', 'D'),
                            ('Виз сзади', 'BACK', 'DU'),
                            ('Вид слева', 'LEFT', 'RL'),
                            ('Вид справа', 'RIGHT', 'R'),
                            ('Вид сверху', 'TOP', 'U'),
                            ('Вид снизу', 'BOTTOM', 'UD'),
                            ('Вид из камеры', 'CAMERA', 'L')):
        def check(self, context, event, action):
            return action == self.ac
        def op(self, context):
            rv3d = context.region_data
            if rv3d and not rv3d.lock_rotation:
                bpy.ops.view3d.viewnumpad(type=self.vtype)
        item = MGItem(name, check, op, vtype=vtype, ac=ac)
        viewitems.append(item)

    def check(self, context, event, action):
        return action == 'LR'
    def op(self, context):
        bpy.ops.view3d.view_persportho('INVOKE_REGION_WIN')
    viewitems.append(MGItem('Вид Персектива/Отография', check, op))

    def check(self, context, event, action):
        return action == 'DR'
    def op(self, context):
        bpy.ops.view3d.localview()
    viewitems.append(MGItem('Вид глобальный/локальный', check, op))

    def check(self, context, event, action):
        return action == 'LU'
    def op(self, context):
        if hasattr(context.space_data, 'use_local_grid'):
            bpy.ops.view3d.view_all(localgrid=True)
        else:
            bpy.ops.view3d.view_all()  # center=True == Ctrl + C
    viewitems.append(MGItem('Вид всё', check, op))

    def check(self, context, event, action):
        return action == 'LD'
    def op(self, context):
        bpy.ops.view3d.view_selected()
    viewitems.append(MGItem('Вид выделенное', check, op))

    def check(self, context, event, action):
        return action.startswith(('LDR', 'ULDR'))
    def op(self, context):
        bpy.ops.view3d.view_center_cursor()
    viewitems.append(MGItem('Центрировать вид на 3D курсор', check, op))

    def check(self, context, event, action):
        return action in ('LDL', 'RD')
    def op(self, context):
        scn = context.scene
        cursor_bak = scn.cursor_location[:]
        bpy.ops.view3d.snap_cursor_to_selected()
        bpy.ops.view3d.view_center_cursor()
        scn.cursor_location = cursor_bak
    viewitems.append(MGItem('Вид выделенное сохранить масштаб', check, op))

    mgdict['view'] = {'items': viewitems, 'relative': False}

    '''### Generic ###
    genitems = []
    # Space Menu (call menu test)
    def check(self, context, event, action):
        return action == 'D'
    def op(self, context):
        bpy.ops.wm.call_menu('INVOKE_REGION_WIN',
                             name='VIEW3D_MT_Space_Dynamic_Menu')
    genitems.append(MGItem('Space Dynamic Menu', check, op))
    mgdict['generic'] = {'items': genitems, 'relative': False}
    '''

    ### Common ###
    commonitems = []
    # help
    def check(self, context, event, action):
        return action in ('UDLU', 'DURD')
    def op(self, context):
        print(help_str)
    commonitems.append(MGItem(help_str, check, op))

    # Mesh Select Overlap
    def check(self, context, event, action):
        if action == 'UDU':
            if hasattr(context.tool_settings, 'use_select_overlap'):
                return True
        return False
    def op(self, context):
        ts = context.tool_settings
        ts.use_select_overlap = not ts.use_select_overlap
    commonitems.append(MGItem('Переключить полисетку выделенныю перекрытием', check, op))

    mgdict['common'] = {'items': commonitems}

    ############################
    ### Change Mouse Gesture ###
    ############################
    chitems = []
    def check(self, context, event, action):
        return action == 'D'
    def op(self, context):
        mg = context.scene.mouse_gesture
        mg.type = self.mgtype
        mg.relative = mgdict[self.mgtype]['relative']
    chitems.append(MGItem('Преобразование', check, op, mgtype='transform'))

    def check(self, context, event, action):
        return action == 'U'
    def op(self, context):
        mg = context.scene.mouse_gesture
        mg.type = self.mgtype
        mg.relative = mgdict[self.mgtype]['relative']
    chitems.append(MGItem('Вид', check, op, mgtype='view'))

    '''def check(self, context, event, action):
        return action == 'R'
    def op(self, context):
        mg = context.scene.mouse_gesture
        mg.type = self.mgtype
        mg.relative = mgdict[self.mgtype]['relative']
    chitems.append(MGItem('Generic', check, op, mgtype='generic'))
    '''

    mgdict['change_mgtype'] = {'items': chitems, 'relative': False}

    return mgdict


mgitems = (('transform', 'Transform', ''),
           ('view', 'View', ''))
default_mgtype = 'transform'  # type: str
### End #######################################################################

mgdict = generate_mgdict()  # type: dict
action_mouse_checked = False


class MouseGesture(bpy.types.PropertyGroup):
    EP = EnumProperty
    txt = 'Gesture "UDLU" or "DURD", ' + \
          'list of mouse gestures appear in the terminal.'
    type = EnumProperty(items=mgitems,
                        name='Mouse Gesture Type',
                        description=txt,
                        default=default_mgtype)
    txt = 'Allow moving the mouse outside the view on some manipulator' + \
          '(transform, ui control drag)'
    do_continuous_grab = BoolProperty(name='Continuous Grab',
                                      description=txt,
                                      default=False)
    relative = BoolProperty(name='Relative',
                            default=False,
                            options={'HIDDEN'})


class VIEW3D_PT_mouse_gesture(bpy.types.Panel):
    bl_space_type = 'VIEW_3D'
    bl_region_type = 'UI'  # 'TOOLS'
    bl_label = 'Жесты мышью'
    bl_options = {'DEFAULT_CLOSED'}

    @classmethod
    def poll(cls, context):
        return context.area is not None

    def draw(self, context):
        mg = context.scene.mouse_gesture
        layout = self.layout
        col = layout.column()
        col.prop(mg, 'type', text='Type')
        if mg.type == 'transform':
            col.prop(mg, 'do_continuous_grab')


class VIEW3D_OT_mouse_gesture(bpy.types.Operator):
    '''Mouse Gesture.'''
    bl_description = 'Mouse gesture.'
    bl_idname = 'view3d.mouse_gesture'
    bl_label = 'Mouse Gesture'
    bl_options = {'REGISTER'}
    # {'REGISTER', 'UNDO', 'BLOCKING', 'MACRO', 'GRAB_POINTER'}

    @classmethod
    def poll(cls, context):
        return context.area.type == 'VIEW_3D'

    def draw_callback_px(self, context):
        if context.region.id != self.region_id:
            return
        posx = 70
        posy1 = 30
        posy2 = 50
        text_interval = 5
        font_id = 0
        blf.size(font_id, 13, context.user_preferences.system.dpi)
        bgl.glEnable(bgl.GL_BLEND)
        if self.changing_mgtype:
            bgl.glColor4f(1.0, 1.0, 0.0, 1.0)
        else:
            bgl.glColor4f(1.0, 1.0, 1.0, 1.0)

        # draw origin
        bgl.glLineWidth(2)
        radius = path_threthold
        radius2 = radius + 5
        x, y, z = self.path[0]
        bgl.glBegin(bgl.GL_LINES)
        for i in range(4):
            r = math.pi / 2 * i + math.pi / 4
            bgl.glVertex2f(x + radius * math.cos(r), y + radius * math.sin(r))
            bgl.glVertex2f(x + radius2 * math.cos(r),
                           y + radius2 * math.sin(r))
        bgl.glEnd()
        bgl.glLineWidth(1)

        # draw lines
        bgl.glEnable(bgl.GL_LINE_STIPPLE)
        bgl.glLineStipple(1, int(0b101010101010101))  # (factor, pattern)
        bgl.glBegin(bgl.GL_LINE_STRIP)
        for v in self.path:
            bgl.glVertex2f(v[0], v[1])
        bgl.glEnd()
        bgl.glLineStipple(1, 1)
        bgl.glDisable(bgl.GL_LINE_STIPPLE)

        # draw txt
        if self.action or self.mgitem:
            x = posx
            for txt in self.action:
                blf.position(font_id, x, posy1, 0)
                result_act = txt
                if txt == 'D':
                    result_act = chr(8595)
                if txt == 'U':
                    result_act = chr(8593)
                if txt == 'R':
                    result_act = chr(8594)
                if txt == 'L':
                    result_act = chr(8592)
                
                blf.draw(font_id, result_act)
                text_width, text_height = blf.dimensions(font_id, txt)
                x += text_width + text_interval
            if self.mgitem:
                countline=0
                linesCount=len(self.mgitem.name.splitlines())
                for line in self.mgitem.name.splitlines():
 
                    blf.position(font_id, posx, posy2+18*linesCount-18-18*countline , 0)
                    blf.draw(font_id, line)
                    countline +=  1
        else:
            #blf.position(font_id, posx, posy2, 0)
            #blf.draw(font_id, '[Mouse Gesture]')
            pass

        # restore opengl defaults
        bgl.glLineWidth(1)
        bgl.glDisable(bgl.GL_BLEND)
        bgl.glColor4f(0.0, 0.0, 0.0, 1.0)
        blf.size(0, 11, context.user_preferences.system.dpi)

    def append_action(self, rvec, vecs):
        angles = [rvec.angle(vec) for vec in vecs]
        a = ('U', 'D', 'R', 'L')[angles.index(min(angles))]
        if not self.action or (self.action and a != self.action[-1]):
            self.action.append(a)

    def get_mgitem(self, context, event, name):
        for mgitem in mgdict[name]['items']:
            if mgitem.check(context, event, ''.join(self.action)):
                break
        else:
            mgitem = None
        return mgitem

    def modal(self, context, event):
        mg = context.scene.mouse_gesture

        if self.op_running_modal:
            # for 'Continuous Grab'
            return {'FINISHED'}

        if context.user_preferences.inputs.select_mouse == 'RIGHT':
            action_mouse = 'LEFTMOUSE'
            select_mouse = 'RIGHTMOUSE'
        else:
            action_mouse = 'RIGHTMOUSE'
            select_mouse = 'LEFTMOUSE'

        mouseco = Vector((event.mouse_region_x, event.mouse_region_y, 0.0))
        mouseco_rel = mouseco - self.path[-1]
        path_length = 0.0
        for i in range(len(self.path) - 1):
            path_length += (self.path[i] - self.path[i + 1]).length

        if event.type == 'MOUSEMOVE' and mouseco_rel.length > path_threthold:
            self.path.append(mouseco)
            if mg.relative:
                if len(self.path) == 1:
                    vec = mouseco - self.path[0]
                    vec.normalize()
                else:
                    vec = self.path[1] - self.path[0]
                    vec.normalize()
                vals = (vec, -vec, (vec[1], -vec[0], 0), (-vec[1], vec[0], 0))
            else:
                vals = ((0, 1, 0), (0, -1, 0), (1, 0, 0), (-1, 0, 0))
            vecs = [Vector(v) for v in vals]  # [U, D, R, L]
            self.append_action(mouseco_rel, vecs)

            if mg.type not in mgdict and not self.changing_mgtype:
                mgitem = None
            else:
                if self.changing_mgtype:
                    mgitem = self.get_mgitem(context, event, 'change_mgtype')
                else:
                    mgitem = self.get_mgitem(context, event, 'common')
                    if mgitem is None:
                        mgitem = self.get_mgitem(context, event, mg.type)
            self.mgitem = mgitem
            context.area.tag_redraw()

        elif (event.type in ('ESC', ) or
              (event.type in (select_mouse, 'MIDDLEMOUSE') and
               event.value == 'PRESS')):
            # cancel
            # context.region.callback_remove(self._handle) --> old 2.66
            bpy.types.SpaceView3D.draw_handler_remove(VIEW3D_OT_mouse_gesture._handle, 'WINDOW') # NEW
            context.area.tag_redraw()
            return {'CANCELLED'}

        elif ((event.type == action_mouse and event.value == 'RELEASE') or
              event.type in ('RET', 'NUMPAD_ENTER') or
              path_length > path_max_length):
            # execute
            # context.region.callback_remove(self._handle) --> old 2.66
            bpy.types.SpaceView3D.draw_handler_remove(VIEW3D_OT_mouse_gesture._handle, 'WINDOW') # NEW
            context.area.tag_redraw()
            if self.mgitem:
                retval = self.mgitem.op(context)
                if retval == {'RUNNING_MODAL'} and mg.do_continuous_grab:
                    self.op_running_modal = 1
                    return {'RUNNING_MODAL'}
            return {'FINISHED'}

        elif event.type in ('SPACE', 'TAB') and event.value == 'PRESS':
            # change mg.type
            mg.relative = mgdict['change_mgtype']['relative']
            self.changing_mgtype = True
            # update action
            self.action = []
            vals = ((0, 1, 0), (0, -1, 0), (1, 0, 0), (-1, 0, 0))
            vecs = [Vector(v) for v in vals]
            for i in range(len(self.path) - 1):
                rvec = self.path[i + 1] - self.path[i]
                self.append_action(rvec, vecs)

            self.mgitem = self.get_mgitem(context, event, 'change_mgtype')
            context.area.tag_redraw()

        return {'RUNNING_MODAL'}

    def invoke(self, context, event):
        global action_mouse_checked
        if not action_mouse_checked:
            km = context.window_manager.keyconfigs.active.keymaps['3D View']
            for kmi in km.keymap_items:
                if kmi.idname == 'view3d.cursor3d':
                    if kmi.type == 'ACTIONMOUSE':
                        kmi.value = 'CLICK'
            action_mouse_checked = True

        if context.area.type != 'VIEW_3D':
            self.report({'WARNING'}, "View3D not found, cannot run operator")
            return {'CANCELLED'}

        context.window_manager.modal_handler_add(self)
        # self._handle = context.region.callback_add(
        #                              self.__class__.draw_callback_px,
        #                              (self, context), 'POST_PIXEL') #--> old 2.66
        VIEW3D_OT_mouse_gesture._handle = bpy.types.SpaceView3D.draw_handler_add(self.__class__.draw_callback_px, (self, context), 'WINDOW', 'POST_PIXEL') # NEW
        self.region_id = context.region.id
        mouseco = Vector((event.mouse_region_x, event.mouse_region_y, 0.0))
        self.path = [mouseco]
        self.action = []
        self.changing_mgtype = False
        for mgitem in mgdict['change_mgtype']['items']:
            if mgitem.mgtype == context.scene.mouse_gesture.type:
                mgitem.op(context)
                break
        mgitem = self.get_mgitem(context, event, 'common')
        if mgitem is None:
            mgitem = self.get_mgitem(context, event,
                                     context.scene.mouse_gesture.type)
        self.mgitem = mgitem

        self.op_running_modal = 0
        context.area.tag_redraw()
        return {'RUNNING_MODAL'}


# Register
def register():
    bpy.utils.register_module(__name__)
    km = bpy.context.window_manager.keyconfigs.active.keymaps['3D View']
    kmi = km.keymap_items.new('view3d.mouse_gesture', 'EVT_TWEAK_A', 'ANY')
    for kmi in km.keymap_items:
        if kmi.idname == 'view3d.cursor3d':
            if kmi.type == 'ACTIONMOUSE':
                kmi.value = 'CLICK'

    bpy.types.Scene.mouse_gesture = \
        PointerProperty(name='Mouse Gesture', type=MouseGesture)


def unregister():
    bpy.utils.unregister_module(__name__)
    km = bpy.context.window_manager.keyconfigs.active.keymaps['3D View']
    for kmi in km.keymap_items:
        if kmi.idname == 'view3d.mouse_gesture':
            km.keymap_items.remove(kmi)
        elif kmi.idname == 'view3d.cursor3d':
            if kmi.type == 'ACTIONMOUSE':
                kmi.value = 'PRESS'

    if bpy.context.scene.get('mouse_gesture') != None:
        del bpy.context.scene['mouse_gesture']
    del bpy.types.Scene.mouse_gesture


if __name__ == '__main__':
    register()